import base64
import re
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, cast

import yaml
from yaml.emitter import Emitter
from yaml.events import AliasEvent, CollectionStartEvent, MappingEndEvent, ScalarEvent, SequenceEndEvent
from yaml.serializer import Serializer

COMMENT_PLACEMENT_BEFORE = "before"
COMMENT_PLACEMENT_INLINE = "inline"

yaml_value_comment_annotation_re = re.compile(r"^__yq_comment_(?P<placement>before|inline)_(?P<key>.+)__$")
yaml_item_comment_annotation_re = re.compile(
    r"^__yq_comment_(?P<placement>before|inline)_(?P<key>\d+)_(?P<value>.+)__$"
)


@dataclass
class YamlComment:
    value: str
    line: int
    inline: bool
    consumed: bool = False


def encode_comment(value: str) -> str:
    return base64.urlsafe_b64encode(value.encode()).decode().rstrip("=")


def decode_comment(value: str) -> str:
    value += "=" * (-len(value) % 4)
    return base64.urlsafe_b64decode(value.encode()).decode()


def make_mapping_comment_key(placement: str, key: str) -> str:
    return "__yq_comment_{}_{}__".format(placement, key)


def make_sequence_comment_annotation(placement: str, index: int, value: str) -> str:
    return "__yq_comment_{}_{}_{}__".format(placement, index, encode_comment(value))


def normalize_comment_values(value: Any) -> List[str]:
    if isinstance(value, list):
        return [str(item) for item in value]
    return [str(value)]


def consume_comments_for_node(loader: Any, anchor_node: Any, value_node: Optional[Any] = None) -> Dict[str, List[str]]:
    result: Dict[str, List[str]] = {COMMENT_PLACEMENT_BEFORE: [], COMMENT_PLACEMENT_INLINE: []}
    comments = getattr(loader, "yaml_comments", [])
    if not comments:
        return result

    inline_nodes = [anchor_node]
    if value_node is not None:
        inline_nodes.append(value_node)
    inline_lines = {node.end_mark.line for node in inline_nodes}

    for comment in comments:
        if not comment.consumed and not comment.inline and comment.line < anchor_node.start_mark.line:
            result[COMMENT_PLACEMENT_BEFORE].append(comment.value)
            comment.consumed = True

    for comment in comments:
        if not comment.consumed and comment.inline and comment.line in inline_lines:
            result[COMMENT_PLACEMENT_INLINE].append(comment.value)
            comment.consumed = True

    return result


class CommentPreservingLoader(yaml.SafeLoader):
    def __init__(self, stream: Any) -> None:
        self.yaml_comments: List[YamlComment] = []
        super().__init__(stream)

    def scan_to_next_token(self) -> None:
        if self.index == 0 and self.peek() == "\ufeff":
            self.forward()
        found = False
        while not found:
            while self.peek() == " ":
                self.forward()
            if self.peek() == "#":
                mark = self.get_mark()
                inline = not self.allow_simple_key
                self.forward()
                chunks = []
                while self.peek() not in "\0\r\n\x85\u2028\u2029":
                    chunks.append(self.peek())
                    self.forward()
                self.yaml_comments.append(YamlComment(value="".join(chunks), line=mark.line, inline=inline))
            if self.scan_line_break():
                if not self.flow_level:
                    self.allow_simple_key = True
            else:
                found = True


def _copy_comment_attrs(event: Any, node: Any) -> Any:
    for attr in "yaml_comment_before", "yaml_comment_inline":
        comments = getattr(node, attr, None)
        if comments:
            setattr(event, attr, comments)
    return event


class CommentPreservingDumperMixin:
    def __getattr__(self, name: str) -> Any:
        raise AttributeError(name)

    def serialize_node(self, node: Any, parent: Any, index: Any) -> None:
        before = getattr(node, "yaml_comment_before", None)
        inline = getattr(node, "yaml_comment_inline", None)
        if not (before or inline):
            Serializer.serialize_node(cast(Serializer, self), node, parent, index)
            return

        dumper = cast(Any, self)
        original_emit = dumper.emit
        attached = False

        def emit_with_comments(event: Any) -> None:
            nonlocal attached
            if not attached and isinstance(event, (AliasEvent, CollectionStartEvent, ScalarEvent)):
                _copy_comment_attrs(event, node)
                attached = True
            original_emit(event)

        dumper.emit = emit_with_comments
        try:
            Serializer.serialize_node(cast(Serializer, self), node, parent, index)
        finally:
            dumper.emit = original_emit

    def expect_node(
        self, root: bool = False, sequence: bool = False, mapping: bool = False, simple_key: bool = False
    ) -> Any:
        if not (root or simple_key) and isinstance(self.event, CollectionStartEvent):
            self.write_inline_comments(getattr(self.event, "yaml_comment_inline", None))
        return Emitter.expect_node(
            cast(Emitter, self), root=root, sequence=sequence, mapping=mapping, simple_key=simple_key
        )

    def expect_block_mapping_key(self, first: bool = False) -> Any:
        if not first and isinstance(self.event, MappingEndEvent):
            return Emitter.expect_block_mapping_key(cast(Emitter, self), first=first)
        self.write_comments_before(getattr(self.event, "yaml_comment_before", None))
        return Emitter.expect_block_mapping_key(cast(Emitter, self), first=first)

    def expect_block_sequence_item(self, first: bool = False) -> Any:
        if not first and isinstance(self.event, SequenceEndEvent):
            return Emitter.expect_block_sequence_item(cast(Emitter, self), first=first)
        self.write_comments_before(getattr(self.event, "yaml_comment_before", None))
        return Emitter.expect_block_sequence_item(cast(Emitter, self), first=first)

    def process_scalar(self) -> None:
        Emitter.process_scalar(cast(Emitter, self))
        self.write_inline_comments(getattr(self.event, "yaml_comment_inline", None))

    def write_comments_before(self, comments: Optional[List[str]]) -> None:
        for comment in comments or []:
            self.write_indent()
            self.write_comment(comment)
            self.write_line_break()

    def write_inline_comments(self, comments: Optional[List[str]]) -> None:
        if not comments:
            return
        self.write_comment(comments[0])
        for comment in comments[1:]:
            self.write_line_break()
            self.write_indent()
            self.write_comment(comment)

    def write_comment(self, comment: str) -> None:
        emitter = cast(Any, self)
        if comment and not comment.startswith(" "):
            comment = " " + comment
        data: Any = "#" + comment
        if emitter.column and not emitter.whitespace:
            data = " " + data
        emitter.whitespace = False
        emitter.indention = False
        emitter.column += len(data)
        emitter.open_ended = False
        if emitter.encoding:
            data = data.encode(emitter.encoding)
        emitter.stream.write(data)



import re
from typing import Any, Dict, List

import yaml

from .loader import hash_key, set_yaml_grammar
from .yaml_support import (
    CommentPreservingDumperMixin,
    decode_comment,
    normalize_comment_values,
    yaml_item_comment_annotation_re,
    yaml_value_comment_annotation_re,
)

# try:
#     from yaml import CSafeDumper as default_dumper
# except ImportError:
#     from yaml import SafeDumper as default_dumper


class OrderedIndentlessDumper(yaml.SafeDumper):
    pass


class OrderedDumper(yaml.SafeDumper):
    def increase_indent(self, flow=False, indentless=False):
        return super(OrderedDumper, self).increase_indent(flow, False)

    def ignore_aliases(self, data):
        return True


class OrderedIndentlessCommentDumper(CommentPreservingDumperMixin, OrderedIndentlessDumper):
    pass


class OrderedCommentDumper(CommentPreservingDumperMixin, OrderedDumper):
    pass


yaml_value_annotation_re = re.compile(r"^__yq_(?P<type>tag|style)_(?P<key>.+)__$")
yaml_item_annotation_re = re.compile(r"^__yq_(?P<type>tag|style)_(?P<key>\d+)_(?P<value>.+)__$")


def get_dumper(use_annotations=False, indentless=False, grammar_version="1.1"):
    # if not (use_annotations or indentless):
    #     return default_dumper

    def represent_dict(dumper, data):
        pairs, custom_styles, custom_tags = [], {}, {}
        custom_comments: Dict[str, Dict[str, List[str]]] = {}
        for k, v in data.items():
            if use_annotations and isinstance(k, str):
                if k == "__yq_alias__":
                    continue
                comment_annotation = yaml_value_comment_annotation_re.match(k)
                if comment_annotation:
                    comment_key = comment_annotation.group("key")
                    placement = comment_annotation.group("placement")
                    custom_comments.setdefault(comment_key, {}).setdefault(placement, []).extend(
                        normalize_comment_values(v)
                    )
                    continue
                value_annotation = yaml_value_annotation_re.match(k)
                if value_annotation and value_annotation.group("type") == "style":
                    custom_styles[value_annotation.group("key")] = v
                    continue
                elif value_annotation and value_annotation.group("type") == "tag":
                    custom_tags[value_annotation.group("key")] = v
                    continue
            pairs.append((k, v))
        mapping = dumper.represent_mapping("tag:yaml.org,2002:map", pairs)
        if use_annotations:
            for k, v in mapping.value:
                hashed_key = hash_key(k.value)
                comments = custom_comments.get(hashed_key, {})
                if "before" in comments:
                    k.yaml_comment_before = comments["before"]
                if "inline" in comments:
                    v.yaml_comment_inline = comments["inline"]
                if hashed_key in custom_styles:
                    if isinstance(v, yaml.nodes.ScalarNode):
                        v.style = custom_styles[hashed_key]
                    elif custom_styles[hashed_key] == "flow":
                        v.flow_style = True
                if hashed_key in custom_tags:
                    v.tag = custom_tags[hashed_key]
        return mapping

    def represent_list(dumper, data):
        raw_list, custom_styles, custom_tags = [], {}, {}
        custom_comments: Dict[str, Dict[str, List[str]]] = {}
        for v in data:
            if use_annotations and isinstance(v, str):
                comment_annotation = yaml_item_comment_annotation_re.match(v)
                if comment_annotation:
                    comment_key = comment_annotation.group("key")
                    placement = comment_annotation.group("placement")
                    comment = decode_comment(comment_annotation.group("value"))
                    custom_comments.setdefault(comment_key, {}).setdefault(placement, []).append(comment)
                    continue
                annotation = yaml_item_annotation_re.match(v)
                if annotation and annotation.group("type") == "style":
                    custom_styles[annotation.group("key")] = annotation.group("value")
                    continue
                elif annotation and annotation.group("type") == "tag":
                    custom_tags[annotation.group("key")] = annotation.group("value")
                    continue
            raw_list.append(v)
        sequence = dumper.represent_list(raw_list)
        if use_annotations:
            for i, v in enumerate(sequence.value):
                item_key = str(i)
                comments = custom_comments.get(item_key, {})
                if "before" in comments:
                    v.yaml_comment_before = comments["before"]
                if "inline" in comments:
                    v.yaml_comment_inline = comments["inline"]
                if item_key in custom_styles:
                    if isinstance(v, yaml.nodes.ScalarNode):
                        v.style = custom_styles[item_key]
                    elif custom_styles[item_key] == "flow":
                        v.flow_style = True
                if item_key in custom_tags:
                    v.tag = custom_tags[item_key]
        return sequence

    dumper: Any
    if use_annotations:
        dumper = OrderedIndentlessCommentDumper if indentless else OrderedCommentDumper
    else:
        dumper = OrderedIndentlessDumper if indentless else OrderedDumper
    dumper.add_representer(dict, represent_dict)
    dumper.add_representer(list, represent_list)
    set_yaml_grammar(dumper, grammar_version=grammar_version)
    return dumper



from datetime import date, datetime, time

import tomlkit
from tomlkit.items import AoT, Array, Bool, Float, InlineTable, Integer, Item, String, Table
from tomlkit.toml_document import TOMLDocument

TOML_META_KEY = "__tomlq_meta__"


def tomlkit_to_json(value, use_annotations=False):
    if isinstance(value, (TOMLDocument, Table, InlineTable)):
        result = {}
        for key, item in value.items():
            result[key] = tomlkit_to_json(item, use_annotations=use_annotations)
        if use_annotations:
            result[TOML_META_KEY] = {"source": _toml_fragment(value), "type": type(value).__name__}
        return result
    if isinstance(value, AoT):
        return [tomlkit_to_json(item, use_annotations=use_annotations) for item in value]
    if isinstance(value, Array):
        return [tomlkit_to_json(item, use_annotations=use_annotations) for item in value]
    if hasattr(value, "unwrap"):
        return value.unwrap()
    return value


def tomlkit_from_json(value):
    if not isinstance(value, dict):
        return value

    meta = value.get(TOML_META_KEY)
    if isinstance(meta, dict) and isinstance(meta.get("source"), str):
        try:
            doc = tomlkit.parse(meta["source"])
        except Exception:
            doc = tomlkit.document()
    else:
        doc = tomlkit.document()

    _apply_mapping(doc, value)
    return doc


def _toml_fragment(value):
    if isinstance(value, InlineTable):
        return tomlkit.dumps(value.unwrap())
    return value.as_string()


def _plain_items(mapping):
    return [(key, value) for key, value in mapping.items() if key != TOML_META_KEY]


def _apply_mapping(container, data):
    desired = {key for key, _ in _plain_items(data)}
    for key in list(container.keys()):
        if key not in desired:
            del container[key]

    for key, value in _plain_items(data):
        if key in container:
            current = container[key]
            replacement = _overlay_item(current, value)
            if replacement is not current:
                container[key] = replacement
        else:
            container[key] = _new_item(value)

    return container


def _overlay_item(current, value):
    if isinstance(value, dict):
        if isinstance(current, (Table, InlineTable, TOMLDocument)):
            return _apply_mapping(current, value)
        return tomlkit_from_json(value)

    if isinstance(value, list):
        if isinstance(current, AoT) and all(isinstance(item, dict) for item in value):
            _apply_aot(current, value)
            return current
        if isinstance(current, Array):
            _apply_array(current, value)
            return current
        return [_new_plain_value(item) for item in value]

    if _json_value(current) == value:
        return current

    return _replacement_item(current, value)


def _apply_array(array, values):
    common_len = min(len(array), len(values))
    for index in range(common_len):
        replacement = _overlay_item(array[index], values[index])
        if replacement is not array[index]:
            array[index] = replacement

    while len(array) > len(values):
        del array[-1]

    for value in values[common_len:]:
        array.append(_new_plain_value(value))


def _apply_aot(aot, values):
    common_len = min(len(aot), len(values))
    for index in range(common_len):
        _apply_mapping(aot[index], values[index])

    while len(aot) > len(values):
        del aot[-1]

    for value in values[common_len:]:
        table = tomlkit.table()
        _apply_mapping(table, value)
        aot.append(table)


def _replacement_item(current, value):
    item: Item
    if isinstance(current, String) and isinstance(value, str):
        try:
            item = String.from_raw(value, current.type)
        except Exception:
            item = tomlkit.item(value)
    elif isinstance(current, Integer) and isinstance(value, int) and not isinstance(value, bool):
        item = tomlkit.integer(value)
    elif isinstance(current, Float) and isinstance(value, (float, int)) and not isinstance(value, bool):
        item = tomlkit.float_(value)
    elif isinstance(current, Bool) and isinstance(value, bool):
        item = tomlkit.boolean(value)
    else:
        item = tomlkit.item(_new_plain_value(value))

    _copy_trivia(current, item)
    return item


def _copy_trivia(source, target):
    if hasattr(source, "trivia") and hasattr(target, "trivia"):
        target.trivia.indent = source.trivia.indent
        target.trivia.comment_ws = source.trivia.comment_ws
        target.trivia.comment = source.trivia.comment
        target.trivia.trail = source.trivia.trail


def _new_item(value):
    value = _new_plain_value(value)
    if isinstance(value, dict):
        table = tomlkit.table()
        _apply_mapping(table, value)
        return table
    if isinstance(value, list):
        return [_new_plain_value(item) for item in value]
    return value


def _new_plain_value(value):
    if isinstance(value, dict):
        return {key: _new_plain_value(item) for key, item in _plain_items(value)}
    if isinstance(value, list):
        return [_new_plain_value(item) for item in value]
    return value


def _json_value(item):
    if hasattr(item, "unwrap"):
        value = item.unwrap()
    else:
        value = item
    if isinstance(value, (datetime, date, time)):
        return value.isoformat()
    return value
